from optparse import OptionParser, OptionGroup, OptionValueError

from . import helpers
from .version import VERSION


def _command_name_to_metavar(command_name):
    """Transforms a command name to a better metaname (used for -h)."""
    return command_name.replace("set_", "").upper()


def _check_color(option, opt_str, value, parser):
    """OptionParser callback to check if the given color is valid."""
    if not helpers.is_color(value):
        raise OptionValueError("option %s: invalid color: '%s'" % (opt_str, value))  # noqa
    setattr(parser.values, option.dest, value)


def _check_colorshift(option, opt_str, values, parser):
    """OptionParser callback to check if the given colors are valid.
    Also transfroms a little bit the args to make it works:
    [color1, color2, speed] -> [[color1, color2], speed]
    """
    colors = values[:-1]
    speed = values[-1]
    for color in colors:
        if not helpers.is_color(color):
            raise OptionValueError("option %s: invalid color: '%s'" % (opt_str, color))  # noqa
    if not speed.isdigit():
        raise OptionValueError("option %s: invalid speed: '%s'" % (opt_str, speed))  # noqa
    setattr(parser.values, option.dest, [colors, int(speed)])


def _check_rgbuniversal(option, opt_str, value, parser):
    """OptionParser callback to check and process inputs for rgbuniversal.
    Accepts 2 formats:
        -x color
        -x time,triggers,color1,pos1[,...,colorn,posn]
    In both cases, it sets option.dest to a 4tuple:
        (colors, positions, time, triggers)
    Note that time and triggers can be "x" for "don't care", and gets assigned
    a default value later on.
    """
    args = value.split(",")
    if len(args) == 1:
        if not helpers.is_color(value):
            raise OptionValueError("option %s: invalid color: '%s'" % (opt_str, value))  # noqa
        setattr(parser.values, option.dest, ([value], ["0"], "x", "x"))

    elif len(args) >= 4 and len(args) & 1 == 0:
        time = args[0]
        if not (time.isdigit() or time.lower() == "x"):
            raise OptionValueError("option %s: invalid time: '%s'" % (opt_str, time))  # noqa

        triggers = args[1]
        if not (helpers.is_hex(triggers) or triggers.lower() == "x"):
            raise OptionValueError("option %s: invalid triggers: '%s'" % (opt_str, triggers))  # noqa

        colors = args[2::2]
        for color in colors:
            if not helpers.is_color(color):
                raise OptionValueError("option %s: invalid color: '%s'" % (opt_str, color))  # noqa

        positions = args[3::2]
        for position in positions:
            if not helpers.is_hex(position):
                raise OptionValueError("option %s: invalid color position: '%s'" % (opt_str, position))  # noqa

        setattr(parser.values, option.dest,
                (colors, positions, time, triggers))
    else:
        raise OptionValueError("option %s: invalid number of values" % (opt_str))  # noqa


def _add_choice_option(group, command_name, command):
    description = "%s (values: %s, default: %s)" % (
            command["description"],
            helpers.choices_to_string(command["choices"]),
            str(command["default"])
            )
    group.add_option(
            *command["cli"],
            help=description,
            dest=command_name,
            choices=helpers.choices_to_list(command["choices"]),
            metavar=_command_name_to_metavar(command_name)
            )


def _add_rgbcolor_option(group, command_name, command):
    description = "%s (e.g. red, #ff0000, ff0000, #f00, f00, default: %s)" % (
            command["description"],
            str(command["default"])
            )
    group.add_option(
            *command["cli"],
            dest=command_name,
            help=description,
            type="string",
            action="callback",
            callback=_check_color,
            metavar=_command_name_to_metavar(command_name)
            )


def _add_rgbcolorshift_option(group, command_name, command):
    description = "%s (e.g. red aqua 200, ff0000 00ffff 200, default: %s %s)" % (  # noqa
            command["description"],
            " ".join(command["default"][0]),
            " ".join((str(v) for v in command["default"][1:]))
            )
    group.add_option(
            *command["cli"],
            dest=command_name,
            help=description,
            type="string",
            action="callback",
            callback=_check_colorshift,
            nargs=3,
            metavar="COLOR1 COLOR2 SPEED"
            )


def _add_rgbuniversal_option(group, command_name, command):
    description = (command["description"] +
                   " (e.g. red, #ff0000, ff0000, #f00, f00). "
                   "If more than one value is specified, "
                   "a color shifting effect is set "
                   "(e.g. x,x,red,0,green,54,blue,54) "
                   "syntax: time(ms),trigger_mask,color1,pos1,...,colorn,posn")
    group.add_option(
            *command["cli"],
            dest=command_name,
            help=description,
            type="string",
            action="callback",
            callback=_check_rgbuniversal,
            metavar=_command_name_to_metavar(command_name)
            )


def _add_range_option(group, command_name, command):
    description = "%s (from %i to %i in increments of %i, default: %i)" % (
            command["description"],
            command["range_min"],
            command["range_max"],
            command["range_increment"],
            command["default"]
            )
    group.add_option(
            *command["cli"],
            dest=command_name,
            help=description,
            metavar=_command_name_to_metavar(command_name),
            choices=[str(i) for i in range(
                command["range_min"],
                command["range_max"] + 1,
                command["range_increment"])]
            )


def _add_hotsbtnmap_option(group, command_name, command):
    description = command["description"]
    group.add_option(
            *command["cli"],
            dest=command_name,
            help=description,
            metavar=_command_name_to_metavar(command_name)
            )


def _add_standard_options(parser):
    parser.add_option(
            "-l", "--list",
            help="print compatible mice and exit",
            action="store_true"
            )


def _add_mouse_options(parser, profile):
    group = OptionGroup(parser, "%s Options" % profile["name"])
    command_names = sorted(profile["commands"].keys())
    for command_name in command_names:
        if command_name == "save":
            continue
        command = profile["commands"][command_name]
        adder_name = "_add_%s_option" % command["value_type"]
        if adder_name not in globals():
            raise Exception("Unable to create a CLI option for value type '%s'" % command["value_type"])  # noqa
        globals()[adder_name](group, command_name, command)
    group.add_option(
            "-r", "--reset",
            help="Reset all options to their factory values",
            action="store_true"
            )
    parser.add_option_group(group)


def generate_cli(profile=None):
    parser = OptionParser(
            "Usage: rivalcfg [options]",
            version=VERSION,
            epilog="Please report any bug on Github: https://github.com/flozz/rivalcfg/issues"  # noqa
            )
    _add_standard_options(parser)
    if profile:
        _add_mouse_options(parser, profile)
    return parser



from .rival import rival

rivaldota2 = {
    "name": "SteelSeries Rival Dota 2 Edition",

    "vendor_id": 0x1038,
    "product_id": 0x1392,
    "interface_number": 0,

    "commands": rival["commands"]

}



from __future__ import print_function

import os
import re
import collections


DEBUG = (
        "RIVALCFG_DEBUG" in os.environ or
        "RIVALCFG_DRY" in os.environ or
        "RIVALCFG_PROFILE" in os.environ or
        "RIVALCFG_DEVICE" in os.environ)

DRY = "RIVALCFG_DRY" in os.environ


MouseId = collections.namedtuple(
        "MouseId", ("vendor_id", "product_id"))


def _get_mouse_id_from_env(env_name):
    if env_name not in os.environ:
        return None
    env_value = os.environ[env_name]
    if not re.match(r"^[0-9a-fA-F]{4}:[0-9a-fA-F]{4}$", env_value):
        raise ValueError("%s='%s' is not a valid mouse identifier" % (
            env_name, env_value))
    vendor_id, product_id = env_value.split(":")
    return MouseId(
            vendor_id=int(vendor_id, 16),
            product_id=int(product_id, 16))


def get_debug_profile():
    """Get the profile to debug from the RIVALCFG_PROFILE environment variable,
    if any.

    This profile should be selected as if the corresponding mouse was
    physically plugged to the computer.
    """
    return _get_mouse_id_from_env("RIVALCFG_PROFILE")


def get_debug_device():
    """Get the profile to debug from the RIVALCFG_DEVICE environment variable,
    if any.

    This device should be selected as the one where the commands will be
    written, regardless of the selected profile. This is usefull to debug a
    mouse that have the same command set than an other one but with a different
    product_id.

    If the RIVALCFG_PROFILE is defined but the RIVALCFG_DEVICE is not, this
    function returns the same output that get_debug_profile()."""
    mouse_id = _get_mouse_id_from_env("RIVALCFG_DEVICE")
    if mouse_id:
        return mouse_id
    return _get_mouse_id_from_env("RIVALCFG_PROFILE")


def log(*args):
    """Logs given arguments to stdout if the debug mode is enabled.

    Arguments:
    *args -- the things to log.
    """
    if not DEBUG:
        return
    print("[DEBUG]", *args)


def log_bytes_hex(message, bytes_):
    """Logs given message and bytes converted into hexadecimal to stdout if
    debug mod is enabled.

    Arguments:
    message -- the message to log
    bytes_ -- the bytes to log as hexadecimal numbers
    """
    if not DEBUG:
        return
    print("[DEBUG]", "%s:" % message, " ".join(["%02X" % int(b) for b in bytes_]))  # noqa



rival = {
    "name": "SteelSeries Rival",

    "vendor_id": 0x1038,
    "product_id": 0x1384,
    "interface_number": 0,

    "commands": {

        # TODO
        "set_sensitivity1": {
            "description": "Set sensitivity preset 1",
            "cli": ["-s", "--sensitivity1"],
            "command": [0x03, 0x01],
            "value_type": "range",
            "range_min": 50,
            "range_max": 6500,
            "range_increment": 50,
            "value_transform": lambda x: int(x / 50),
            "default": 800,
        },

        # TODO
        "set_sensitivity2": {
            "description": "Set sensitivity preset 2",
            "cli": ["-S", "--sensitivity2"],
            "command": [0x03, 0x02],
            "value_type": "range",
            "range_min": 50,
            "range_max": 6500,
            "range_increment": 50,
            "value_transform": lambda x: int(x / 50),
            "default": 1600,
        },

        "set_polling_rate": {
            "description": "Set polling rate in Hz",
            "cli": ["-p", "--polling-rate"],
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "set_logo_light_effect": {
            "description": "Set the logo light effect",
            "cli": ["-e", "--logo-light-effect"],
            "command": [0x07, 0x01],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
            },
            "default": "steady",
        },

        "set_wheel_light_effect": {
            "description": "Set the wheel light effect",
            "cli": ["-E", "--wheel-light-effect"],
            "command": [0x07, 0x02],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
            },
            "default": "steady",
        },

        "set_logo_color": {
            "description": "Set the logo backlight color",
            "cli": ["-c", "--logo-color"],
            "command": [0x08, 0x01],
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "set_wheel_color": {
            "description": "Set the wheel backlight color",
            "cli": ["-C", "--wheel-color"],
            "command": [0x08, 0x02],
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "save": {
            "description": "Save the configuration to the mouse memory",
            "cli": None,
            "command": [0x09, 0x00],
            "value_type": None,
        },

    },

}



import rivalcfg


class TestListSupportedMice(object):

    def test_returns_the_supported_mice(self):
        assert len(list(rivalcfg.list_supported_mice())) > 0

    def test_returns_name_vendorid_productid_of_each_mouse(self):
        for mouse in rivalcfg.list_supported_mice():
            assert len(mouse) == 3
            assert mouse.name
            assert mouse.product_id
            assert mouse.vendor_id


class TestListAvailableMice(object):

    def test_not_available_mouse(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:0001")
        assert len(list(rivalcfg.list_available_mice())) == 0

    def test_debug_available_mouse(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1710")
        assert len(list(rivalcfg.list_available_mice())) > 0


class TestGetMouseProfile(object):

    def test_supported_mouse(self):
        assert rivalcfg.get_mouse_profile(0x1038, 0x1710)

    def test_unsupported_mouse(self):
        assert rivalcfg.get_mouse_profile(0x0000, 0x0000) is None


class TestGetMouse(object):

    def test_unsuported_mouse(self):
        assert not rivalcfg.get_mouse(0x1038, 0x0001)

    def test_supported_mouse(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setattr("rivalcfg.debug.DRY", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1710")
        assert rivalcfg.get_mouse(0x1038, 0x1710)


class TestGetFirstMouse(object):

    def test_unsuported_mouse(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setattr("rivalcfg.debug.DRY", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:0001")
        assert not rivalcfg.get_first_mouse()

    def test_supported_mouse(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setattr("rivalcfg.debug.DRY", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1710")
        assert rivalcfg.get_first_mouse()



import pytest

import rivalcfg.mouse


PROFILE = {
    "name": "Fake Mouse",

    "vendor_id": 0x1038,
    "product_id": 0x0001,
    "interface_number": 0,

    "commands": {
        "choice_command": {
            "command": [0x01],
            "value_type": "choice",
            "choices": {
                1: 2,
                2: 4
            }
        },
        "rgbcolor_command": {
            "command": [0x02],
            "value_type": "rgbcolor",
        },
        "range_command": {
            "command": [0x03],
            "value_type": "range",
            "range_min": 4,
            "range_max": 10,
            "range_increment": 2
        },
        "none_command": {
            "command": [0x01, 0x02, 0x03],
            "value_type": None
        },
        "undefined_handler_command": {
            "command": [0x04],
            "value_type": "foo"
        }
    }
}


@pytest.fixture(scope="function")
def mouse(monkeypatch):
    monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
    monkeypatch.setattr("rivalcfg.debug.DRY", True)
    monkeypatch.setenv("RIVALCFG_PROFILE", "1038:0001")
    return rivalcfg.mouse.Mouse(PROFILE)


def _get_written_bytes(mouse):
    mouse._device.seek(0)
    result = mouse._device.read()
    if result[0:1] == b"\x00":
        result = result[1:]
    return result


class TestGetattr(object):

    def test_choice_handler(self, mouse):
        mouse.choice_command(2)
        assert _get_written_bytes(mouse) == b"\x01\x04"

    def test_rgbcolor_handler(self, mouse):
        mouse.rgbcolor_command(0xFF, 0x18, 0x00)
        assert _get_written_bytes(mouse) == b"\x02\xFF\x18\x00"

    def test_range_handler(self, mouse):
        mouse.range_command(10)
        assert _get_written_bytes(mouse) == b"\x03\x0A"

    def test_none_hanlder(self, mouse):
        mouse.none_command()
        assert _get_written_bytes(mouse) == b"\x01\x02\x03"

    def test_undefined_method(self, mouse):
        with pytest.raises(AttributeError):
            mouse.foo_command()

    def test_undefined_handler(self, mouse):
        with pytest.raises(Exception):
            mouse.undefined_handler_command()


class TestStrRepr(object):

    def test_str(self, mouse):
        assert str(mouse) == "<Mouse Fake Mouse (1038:0001:00)>"

    def test_repr(self, mouse):
        assert repr(mouse) == "<Mouse Fake Mouse (1038:0001:00)>"




from .. import usbhid

rival710 = {
    "name": "SteelSeries Rival 710 (Experimental)",

    "vendor_id": 0x1038,
    "product_id": 0x1730,
    "interface_number": 0,

    "commands": {

        "set_sensitivity1": {
            "description": "Set sensitivity preset 1",
            "cli": ["-s", "--sensitivity1"],
            "command": [0x03, 0x00, 0x01],
            "suffix": [0x00, 0x42],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int((x / 100) - 1),
            "default": 800,
        },

        "set_sensitivity2": {
            "description": "Set sensitivity preset 2",
            "cli": ["-S", "--sensitivity2"],
            "command": [0x03, 0x00, 0x02],
            "suffix": [0x00, 0x42],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int((x / 100) - 1),
            "default": 1600,
        },

        "set_logo_color": {
            "description": "Set the logo backlight color",
            "cli": ["-c", "--logo-color"],
            "command": [0x05, 0x00, 0x00],
            "suffix": [0xFF, 0x32, 0xC8, 0xC8, 0x00, 0x00, 0x01],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,  # wValue = 0x0300
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "set_wheel_color": {
            "description": "Set the wheel backlight color",
            "cli": ["-C", "--wheel-color"],
            "command": [0x05, 0x00, 0x01],
            "suffix": [0xFF, 0x32, 0xC8, 0xC8, 0x00, 0x01, 0x01],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,  # wValue = 0x0300
            "value_type": "rgbcolor",
            "default": "#FF1800"
        },

        "save": {
            "description": "Save the configuration to the mouse memory",
            "cli": None,
            "command": [0x09, 0x00],
            "value_type": None,
        },

    },

}



from .. import usbhid

sensei310 = {
    "name": "SteelSeries Sensei 310 (Experimental)",

    "vendor_id": 0x1038,
    "product_id": 0x1722,
    "interface_number": 0,

    "rgbuniversal_format": {
        "header_len": 26,   # Number of bytes in header excluding command bytes
        "led_id": [0],      # Index(es) of LED ID
        "speed": 1,         # Index of the colorshift speed field
        "speed_len": 2,     # How many bytes the speed field takes up
        "repeat": 17,       # Index of the repeat flag
        "triggers": 21,     # Index of the trigger button mask field
        "point_count": 25,  # Index of the color count field
    },

    "commands": {

        "set_sensitivity1": {
            "description": "Set sensitivity preset 1",
            "cli": ["-s", "--sensitivity1"],
            "command": [0x53, 0x00, 0x01],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int(x / 100) - 1,
            "suffix": [0x00, 0x42],
            "default": 800,
        },

        "set_sensitivity2": {
            "description": "Set sensitivity preset 2",
            "cli": ["-S", "--sensitivity2"],
            "command": [0x53, 0x00, 0x02],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int(x / 100) - 1,
            "suffix": [0x00, 0x42],
            "default": 1600,
        },

        "set_polling_rate": {
            "description": "Set polling rate in Hz",
            "cli": ["-p", "--polling-rate"],
            "command": [0x54, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "set_logo_color": {
            "description": "Set the logo backlight color(s) and effects",
            "cli": ["-c", "--logo-color"],
            "command": [0x5B, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x00,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_wheel_color": {
            "description": "Set the wheel backlight color(s) and effects",
            "cli": ["-C", "--wheel-color"],
            "command": [0x5B, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x01,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "save": {
            "description": "Save the configuration to the mouse memory",
            "cli": None,
            "command": [0x59, 0x00],
            "value_type": None,
        },

    },

}



import pytest

import rivalcfg.debug


class TestGetDebugProfil(object):

    def test_env_variable_not_defined(self, monkeypatch):
        assert rivalcfg.debug.get_debug_profile() is None

    def test_env_variable_defined(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        debugMouseInfo = rivalcfg.debug.get_debug_profile()
        assert debugMouseInfo.vendor_id == 0x1038
        assert debugMouseInfo.product_id == 0x1702

    def test_env_variable_defined_to_wrong_value(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_PROFILE", "foobar")
        with pytest.raises(ValueError):
            rivalcfg.debug.get_debug_profile()


class TestGetDebugDevice(object):

    def test_env_variable_not_defined(self, monkeypatch):
        assert rivalcfg.debug.get_debug_device() is None

    def test_env_variable_defined(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DEVICE", "1038:1702")
        debugMouseInfo = rivalcfg.debug.get_debug_device()
        assert debugMouseInfo.vendor_id == 0x1038
        assert debugMouseInfo.product_id == 0x1702

    def test_env_variable_defined_to_wrong_value(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DEVICE", "foobar")
        with pytest.raises(ValueError):
            rivalcfg.debug.get_debug_device()

    def test_device_env_variable_not_defined_but_profile_env_variable_defined(self, monkeypatch):  # noqa
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        debugMouseInfo = rivalcfg.debug.get_debug_device()
        assert debugMouseInfo.vendor_id == 0x1038
        assert debugMouseInfo.product_id == 0x1702

    def test_device_env_variable_and_profile_env_variable_defined(self, monkeypatch):  # noqa
        monkeypatch.setenv("RIVALCFG_DEVICE", "1038:1710")
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        debugMouseInfo = rivalcfg.debug.get_debug_device()
        assert debugMouseInfo.vendor_id == 0x1038
        assert debugMouseInfo.product_id == 0x1710



import pytest

import rivalcfg.usbhid


class TestIsDevicePlugged(object):

    def test_not_plugged_mouse(self):
        assert not rivalcfg.usbhid.is_device_plugged(0x1038, 0x0001)

    def test_debug_mouse_plugged(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:0002")
        assert not rivalcfg.usbhid.is_device_plugged(0x1038, 0x0001)
        assert rivalcfg.usbhid.is_device_plugged(0x1038, 0x0002)


class TestOpenDevice(object):

    def test_not_plugged_mouse(self):
        with pytest.raises(IOError):
            rivalcfg.usbhid.open_device(0x1038, 0x0001, 0x00)

    def test_debug_mouse_plugged(self, monkeypatch):
        monkeypatch.setattr("rivalcfg.debug.DEBUG", True)
        monkeypatch.setattr("rivalcfg.debug.DRY", True)
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:0002")
        device = rivalcfg.usbhid.open_device(0x1038, 0x0002, 0x00)
        assert hasattr(device, "write")
        assert hasattr(device, "close")
        assert hasattr(device, "send_feature_report")
